/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.modules.coverage.data.local;

import com.vladium.emma.EMMARuntimeException;
import com.vladium.emma.IAppErrorCodes;
import com.vladium.emma.data.ICoverageData;
import com.vladium.emma.data.IMetaData;
import com.vladium.emma.report.AbstractReportGenerator;
import com.vladium.emma.report.AllItem;
import com.vladium.emma.report.ClassItem;
import com.vladium.emma.report.IItem;
import com.vladium.emma.report.ItemComparator;
import com.vladium.emma.report.PackageItem;
import com.vladium.emma.report.SourcePathCache;
import com.vladium.emma.report.SrcFileItem;
import com.vladium.util.Descriptors;
import com.vladium.util.IProperties;
import com.vladium.util.IntObjectMap;
import java.util.Iterator;
import java.util.LinkedList;
import org.netbeans.modules.coverage.utils.LineCoverageStates;


/**
 * The class is used to parse EMMA coverage data and produce our own coverage data object
 * The implementation is based on the EMMA's AbstractReportGnerator. See EMMA's HTML report generator for
 * more details
 * @author mikk@netbeans.org
 */
public final class ReportGenerator extends AbstractReportGenerator
        implements IAppErrorCodes {
    
    
    private LinkedList /* IITem */ m_queue;
    private static final String TYPE = "html";
    private static final int IO_BUF_SIZE = 32 * 1024;
    private CoverageData data = null;    
    
    /**
     * Creates the generator
     * @param coverageData Coverage data object to save coverage info to
     */
    public ReportGenerator(CoverageData coverageData){
        this.data = coverageData;
    }
    
    public final String getType() {
        return TYPE;
    }
    
    public void process(final IMetaData mdata, final ICoverageData cdata,
            final SourcePathCache cache, final IProperties properties)
            throws EMMARuntimeException {
        
        initialize(mdata, cdata, cache, properties);
        
        m_queue = new LinkedList();
        
        //System.err.println("Started parsing metadata: "+new Date());
        for (m_queue.add(m_view.getRoot()); ! m_queue.isEmpty(); ) {
            final IItem head = (IItem) m_queue.removeFirst();
            head.accept(this, null);
        }
        //System.err.println("Finished parsing metadata: "+new Date());
    }
    
    
    public void cleanup() {
        m_queue = null;
        super.cleanup();
    }
    
    
    // IItemVisitor:
    public Object visit(final AllItem item, final Object ctx) {
        //we do nothing here except adding child items to queue
        final boolean deeper = (m_settings.getDepth() > item.getMetadata().getTypeID());
        final ItemComparator order = m_typeSortComparators [PackageItem.getTypeMetadata().getTypeID()];
        for (Iterator packages = item.getChildren(order); packages.hasNext();) {
            final IItem pkg = (IItem) packages.next();
            if (deeper) m_queue.addLast(pkg);
        }
        return ctx;
    }
    
    
    public Object visit(final PackageItem item, final Object ctx) {
        //we do nothing here except adding child items to queue
        final boolean deeper = (m_settings.getDepth() > item.getMetadata().getTypeID());
        final ItemComparator order = m_typeSortComparators [m_srcView ? SrcFileItem.getTypeMetadata().getTypeID() : ClassItem.getTypeMetadata().getTypeID()];
        for (Iterator srcORclsFiles = item.getChildren(order); srcORclsFiles.hasNext(); ) {
            final IItem srcORcls = (IItem) srcORclsFiles.next();
            if (deeper) m_queue.addLast(srcORcls);
        }
        return ctx;
    }
    
    
    public Object visit(final SrcFileItem item, final Object ctx) {
        // Information about source file is extracted here
        try {
            final boolean deeper = (m_settings.getDepth() > ClassItem.getTypeMetadata().getTypeID());
            // embed source file:
            if (deeper) {
                handleSourceFile(item);
            }
        } finally {
        }
        return ctx;
    }
    
    
    public Object visit(final ClassItem item, final Object ctx) {
        // this visit only takes place in class views, we do not deal with them
        //we do nothing here
        return ctx;
    }
    
    
    private void handleSourceFile(final SrcFileItem item) {
        
        if (item.getLineCoverage().size()==0){
            //nothing is covered we return;
            return;
        }
        
        ClassCoverageData classCoverage = new ClassCoverageData();
        final String fileName = item.getName();
        final String packageVMName = ((PackageItem) item.getParent()).getVMName();
        
        
        String curClassName = Descriptors.combineVMName(packageVMName, fileName);
        int dotIndex = curClassName.indexOf('.');
        curClassName = curClassName.substring(0, dotIndex).replace('/', '.');
        
        try {
            
            final boolean markupCoverage = m_hasLineNumberInfo;
            final int unitsType = m_settings.getUnitsType();
            IntObjectMap /* line num:int -> SrcFileItem.LineCoverageData */ lineCoverageMap = null;
            
            if (markupCoverage){
                lineCoverageMap = item.getLineCoverage();
            }else{
                //the code has no line number info. We exit
                return;
            }
            
            int[] lineNumbers = lineCoverageMap.keys();
            
            for(int i=0;i<lineNumbers.length;i++){
                int l = lineNumbers[i];
                final SrcFileItem.LineCoverageData lCoverageData = (SrcFileItem.LineCoverageData) lineCoverageMap.get(l);
                
                if (lCoverageData != null) {
                    switch (lCoverageData.m_coverageStatus) {
                        case SrcFileItem.LineCoverageData.LINE_COVERAGE_ZERO:
                            //doNothing
                            break;
                        case SrcFileItem.LineCoverageData.LINE_COVERAGE_PARTIAL:
                        {
                            classCoverage.putLineCoverage(l-1, LineCoverageStates.LINE_COVERAGE_PARTIAL);                            
                        }
                        break;
                        
                        case SrcFileItem.LineCoverageData.LINE_COVERAGE_COMPLETE:
                        {                            
                            classCoverage.putLineCoverage(l-1, LineCoverageStates.LINE_COVERAGE_COMPLETE);
                        }
                        break;
                        
                        
                    } // end of switch
                }
                
            }
            
        } catch (Throwable t) {
            t.printStackTrace(System.out); // TODO: logging
        }
        
        data.putClassCoverage(curClassName, classCoverage);
        
    }
    
} 