/*
 * The contents of this file are subject to the terms of the Common Development
 * and Distribution License (the License). You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the License at http://www.netbeans.org/cddl.html
 * or http://www.netbeans.org/cddl.txt.
 * 
 * When distributing Covered Code, include this CDDL Header Notice in each file
 * and include the License file at http://www.netbeans.org/cddl.txt.
 * If applicable, add the following below the CDDL Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * PatternOperator.java
 */

package org.netbeans.orchlib.core.operators;

import java.awt.Component;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import org.netbeans.orchlib.exceptions.ElementNotFoundException;
import org.netbeans.jemmy.EventTool;
import org.netbeans.jemmy.JemmyProperties;
import org.netbeans.jemmy.Timeout;
import org.netbeans.jemmy.Timeouts;
import org.netbeans.jemmy.Waitable;
import org.netbeans.jemmy.Waiter;
import org.netbeans.jemmy.operators.JPopupMenuOperator;
import org.netbeans.jemmy.operators.Operator;
import org.netbeans.modules.bpel.design.geometry.FPoint;
import org.netbeans.modules.bpel.design.geometry.FRectangle;
import org.netbeans.modules.bpel.design.model.elements.VisualElement;
import org.netbeans.modules.bpel.design.model.patterns.CompositePattern;
import org.netbeans.modules.bpel.design.model.patterns.Pattern;

/**
 * This operator represents the patterns (==activities in most of the cases) which are interactive
 * elements on the diagram
 * 
 * The operator provides ways to correctly find and select the patterns and also to
 * get access to it's internals
 *
 * @author Mikhail.Kondratyev@sun.com
 */
public class PatternOperator extends Operator implements IActionable{
    
    private DesignViewOperator viewOperator = null;
    private Pattern pattern = null;
    
    
    
    public final static long WAIT_PATTERN_TIMEOUT = 60000;
    
    
    public PatternOperator(DesignViewOperator viewOperator) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeChooser(PatternTypes.ANY), 0);
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, int index) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeChooser(PatternTypes.ANY), index);
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, PatternTypes type, int index) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeChooser(type), index);
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, String patternName, PatternTypes type, int index) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeAndNameChooser(type, patternName, getDefaultStringComparator()), index);
    }
    
    public PatternOperator(DesignViewOperator viewOperator, String patternName, PatternTypes type, StringComparator comparator) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeAndNameChooser(type, patternName, comparator), 0);
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, String patternName, int index) throws ElementNotFoundException {
        this(viewOperator, new PatternByTypeAndNameChooser(PatternTypes.ANY, patternName, getDefaultStringComparator()), index);
    }
    
    public PatternOperator(PatternOperator patternOper, String patternName, int index) throws ElementNotFoundException {
        this(patternOper, new PatternByTypeAndNameChooser(PatternTypes.ANY, patternName, getDefaultStringComparator()), index);
    }
    
    public PatternOperator(PatternOperator patternOper, PatternTypes type, int index) throws ElementNotFoundException {
        this(patternOper, new PatternByTypeChooser(type), index);
    }
    
    public PatternOperator(PatternOperator patternOper,PatternTypes type , String patternName, int index) throws ElementNotFoundException {
        this(patternOper, new PatternByTypeAndNameChooser(type, patternName, getDefaultStringComparator()), index);
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, IPatternChooser patternChooser, int index) throws ElementNotFoundException {
        this(viewOperator, waitForPattern(viewOperator.getRootPatternOperator(), patternChooser, index));
        
    }
    
    
    public PatternOperator(PatternOperator patternOper, IPatternChooser patternChooser, int index) throws ElementNotFoundException {
        this(patternOper.getDesignView(), waitForPattern(patternOper, patternChooser, index));
        
    }
    
    
    public PatternOperator(DesignViewOperator viewOperator, Pattern pattern) {
        this.viewOperator = viewOperator;
        this.pattern = pattern;
        
    }
    
    private static Pattern findPattern(PatternOperator patternOperator, IPatternChooser patternChooser, int index) throws ElementNotFoundException{
        if (! (patternOperator.pattern instanceof CompositePattern)){
            throw new ElementNotFoundException();
        }
        
        CompositePattern rootPattern = (CompositePattern)patternOperator.pattern;
        
        ArrayList<Pattern> patternsFound = new ArrayList<Pattern>();
        ArrayList<Pattern> allPatterns = new ArrayList<Pattern>();
                
        /*
        //this is to handle concept of root pattern and allow
        //search of process in design view
        //in fact this is not the best way to handle it but we decided not to rewrite the tests
         */
        //checking if the pattern is process pattern
        if (rootPattern.getParent()==null){
            //for the topmost pattern we allow to find itself inside its children,
            //this is needed to allow search of B.Process inside Design view
            allPatterns.add(rootPattern);
        }
        
        listChildPatterns(allPatterns, rootPattern);
        
        for(int i=0;i<allPatterns.size();i++){
            if (patternChooser.checkPattern(allPatterns.get(i))){
                patternsFound.add(allPatterns.get(i));
            }
        }
        
        //sorting found patterns
        PatternComparator<Pattern> c = new PatternComparator<Pattern>();
        Pattern[] arr = (Pattern[])patternsFound.toArray(new Pattern[0]);
        Arrays.sort(arr,c);
        if (arr.length>index){
            return arr[index];
        }else{
            throw new ElementNotFoundException();
        }
    }
    
    
    private static void listChildPatterns(ArrayList<Pattern> childPatterns, Pattern rootPattern){
        if (rootPattern instanceof CompositePattern){
            CompositePattern composite = (CompositePattern)rootPattern;
            Collection<Pattern> nestedPatterns = composite.getNestedPatterns();
            for (Iterator<Pattern> iter = nestedPatterns.iterator(); iter.hasNext();) {
                Pattern nested = iter.next();
                childPatterns.add(nested);
                listChildPatterns(childPatterns, nested);
            }
        }
    }
    
    
    public static Pattern waitForPattern(final PatternOperator rootPattern, final IPatternChooser patternChooser, final int index) {
        try{
            Waiter w = new Waiter(new Waitable() {
                public Object actionProduced(Object obj) {
                    try {
                        Pattern go = findPattern(rootPattern, patternChooser, index);
                        return go;
                    }catch(ElementNotFoundException e) {
                        return null;
                    }
                }
                public String getDescription() {
                    return("Wait for " + patternChooser.getDescription());
                }
            });
            Timeouts t = JemmyProperties.getCurrentTimeouts();
            t.setTimeout("Waiter.WaitingTime", t.getTimeout("PatternOperator.WaitPatternTime"));
            return (Pattern)w.waitAction(null);
            
        }catch(InterruptedException ie) {
            return null;
        }
    }
    
    
    public PatternOperator getDirectChildPattern(PatternTypes type, int index) throws ElementNotFoundException{
        return getDirectChildPattern(new PatternByTypeChooser(type),  index);
    }
    
    public PatternOperator getDirectChildPattern(IPatternChooser chooser, int index) throws ElementNotFoundException{
        if (! (pattern instanceof CompositePattern)){
            throw new ElementNotFoundException();
        }
        
        ArrayList<Pattern> patternsFound = new ArrayList<Pattern>();
        Collection<Pattern> nested = ((CompositePattern)pattern).getNestedPatterns();
        for (Iterator<Pattern> iter = nested.iterator(); iter.hasNext();) {
            Pattern element = iter.next();
            if (chooser.checkPattern(element)){
                patternsFound.add(element);
            }
        }
        
        //sorting found patterns
        PatternComparator<Pattern> c = new PatternComparator<Pattern>();
        Pattern[] arr = (Pattern[])patternsFound.toArray(new Pattern[0]);
        Arrays.sort(arr,c);
        if (arr.length>index){
            return new PatternOperator(getDesignView(), arr[index]);
        }else{
            throw new ElementNotFoundException();
        }
    }
    
    public int getChildPatternsCount(){
        int res = 0;
        if (this.pattern instanceof CompositePattern){
            res = ((CompositePattern)pattern).getNestedPatterns().size();
        }
        return res;
    }
    
    public int getChildPatternsCount(IPatternChooser chooser){
        int res = 0;
        if (this.pattern instanceof CompositePattern){
            Collection<Pattern> nested = ((CompositePattern)pattern).getNestedPatterns();
            for (Iterator<Pattern> iter = nested.iterator(); iter.hasNext();) {
                Pattern element = iter.next();
                if (chooser.checkPattern(element)){
                    res++;
                }
            }
        }
        return res;
    }
    
    public int getChildPatternsCount(PatternTypes type){
        return getChildPatternsCount(new PatternByTypeChooser(type));
    }
    
        
    public ArrayList<VisualElementOperator> getDirectChildElements(IElementChooser chooser){
        ArrayList<VisualElementOperator> childElements = new ArrayList<VisualElementOperator>();
        Collection<VisualElement> elements = pattern.getElements();
        for (Iterator<VisualElement> iter = elements.iterator(); iter.hasNext();) {
            VisualElement element = iter.next();
            if (chooser.checkElement(element)){
                childElements.add(new VisualElementOperator(element));
            }
        }
        return childElements;
    }
    
    
    public ArrayList<VisualElementOperator> getDirectChildElements(ElementTypes elementType){
        return getDirectChildElements(new VisualElementOperator.ElementByTypeChooser(elementType));
    }
    
    public int getAllChildPatternsCount(){
        ArrayList<Pattern> allPatterns = new ArrayList<Pattern>();
        listChildPatterns(allPatterns, pattern);
        return allPatterns.size();
    }
    
    public int getAllChildPatternsCount(PatternTypes childPatternType){
        return getAllChildPatternsCount(new PatternByTypeChooser(childPatternType));
    }
    
    public int getAllChildPatternsCount(IPatternChooser chooser){
        ArrayList<Pattern> allPatterns = new ArrayList<Pattern>();
        listChildPatterns(allPatterns, pattern);
        int count = 0;
        for(int i=0;i<allPatterns.size();i++){
            if (chooser.checkPattern(allPatterns.get(i))){
                count++;
            }
        }
        return count;
    }
    
    public int getChildElementsCount(){
        return pattern.getElements().size();
    }
    
    public int getAllChildElementsCount(){
        ArrayList<VisualElement> allElements = new ArrayList<VisualElement>();
        VisualElementOperator.listChildElements(allElements, pattern);
        return allElements.size();
        
    }
    
    /**
     * We must implement this method to extend Operator, but since
     * we are not working with Component this method always return null;
     * @return null
     */
    public Component getSource() {
        return null;
    }
    
    Pattern getPattern(){
        return pattern;
    }
    
    public PatternOperator getParent(){
        return (pattern.getParent() != null)?new PatternOperator(viewOperator, pattern.getParent()):null;
    }
    
    public String getName(){        
        VisualElement namedEl = pattern.getTextElement();
        if (namedEl!=null){
            return namedEl.getLabelText();
        }else{
            return null;
        }
    }

    public FPoint getNameCenterPoint(){
        VisualElement namedEl = pattern.getTextElement();
        if (namedEl!=null){
            return namedEl.getTextBounds().getCenter();
        }else{
            return null;
        }
    }
    
    /**
     * Return DesignView, where this element is placed
     */
    public DesignViewOperator getDesignView() {
        return viewOperator;
    }
    
    
    /**
     * returns pattern type
     */
    public String getTypeString() {
        return pattern.getOMReference().getClass().getName();
    }
    
    
    /**
     * @returns pattern type as Enum value
     */
    public PatternTypes getType() {
        String strType = getTypeString();
        getOutput().printLine("=============Searching for enum value for pattern type "+strType);
        PatternTypes type = PatternTypes.ANY;
        for(PatternTypes _type: PatternTypes.values()){
            if(_type.toString().equals(strType)
            && !_type.name().startsWith("SWITCH")){// TODO: remove SWITCH
                type = _type;
                getOutput().printLine("=============Found type for "+strType+", it's "+type.name());
                break;
            }
        }
        return type;
    }
    
    /**
     *
     * @return bounding rect for element
     */
    public Rectangle getBoundingRect() {
        Rectangle rect = pattern.createSelection().getBounds();
        FRectangle frect = new FRectangle(rect.x, rect.y, rect.width, rect.height);
        return getDesignView().convertRectToScreen(frect);
    }
    
    
    public VisualElementOperator getLastElement(){
        VisualElement el = null;
        el = pattern.getLastElement();
        if (el==null){
            return null;
        }
        return new VisualElementOperator(el);
        
    }
    
    
    public VisualElementOperator getFirstElement(){
        VisualElement el = null;
        el = pattern.getFirstElement();
        if (el==null){
            return null;
        }
        return new VisualElementOperator(el);
        
    }
    
    
    public void clickActionPoint(int clickCount, int mouseBtn, int modifiers){
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, clickCount, mouseBtn, modifiers);
    }
    
    public void clickActionPoint(int clickCount){
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, clickCount, InputEvent.BUTTON1_MASK, 0);
    }
    
    
    public void clickForPopup() {
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, 1, InputEvent.BUTTON3_MASK, 0);
    }
    
    public void goToSource(){
        clickForPopup();
        new JPopupMenuOperator().pushMenu("Go To Source");
    }
    
    public SourceElementOperator goToSourceElement()throws ElementNotFoundException{
        goToSource();
        //create SourceElementOperator
        return new SourceElementOperator(new SourceViewOperator(viewOperator.getName()), getType());
    }
    
    public Point getActionPoint(){
        Rectangle rect = getBoundingRect();
        Point initialPoint = rect.getLocation();
//        initialPoint.x+=rect.width/2;
//        initialPoint.y+=2;
        initialPoint.x += rect.width - 2;
        initialPoint.y += rect.height/2;
        return initialPoint;
    }
    
    public ConnectionOperator getConnection() throws ElementNotFoundException{
        VisualElementOperator element1 = null;
        VisualElementOperator element2 = null;
        if(getAllChildElementsCount() == 0){
            element1 = new VisualElementOperator(this);
            element2 = element1;
        }
        else{
            element1 = new VisualElementOperator(this);
            element2 = new VisualElementOperator(this, 1);
        }
        if(element1 == null || element2 == null)
            throw new ElementNotFoundException();
        
        return new ConnectionOperator(element1, element2);
    }
    
    
    public void makeVisible(){
        Rectangle rect = null;
        //int inflate = 15;
        
        Rectangle boundingRect = getBoundingRect();
        //TODO: this comparison is kind of hack and should be work arounded if time permits
        if (boundingRect.getLocation().x<0 || boundingRect.getLocation().y<0){
            rect = new Rectangle(0,0, (int)boundingRect.getMaxX(), (int)boundingRect.getMaxY());
        }else{
            rect = boundingRect;
        }
        
        getDesignView().scrollToPoint(getActionPoint());
        getDesignView().scrollRectToVisible(rect);
        new EventTool().waitNoEvent(500);
    }
    
    
    //Methods from Actionable interface
    
    public JPopupMenuOperator getPopup() {
        clickForPopup();
        return new JPopupMenuOperator();
    }
    
    
    
    public void select() {
        //TODO: wait selected should be added
        clickActionPoint(1, InputEvent.BUTTON1_MASK, 0);
        new Timeout("", 50).sleep();
    }
    
    
    public boolean isSelected(){
        return getDesignView().getDesignView().getSelectionModel().getSelectedPattern()==this.getPattern();
    }
    
    
    
    public void addToSelection() {
        new Timeout("",10).sleep();
        clickActionPoint(1, InputEvent.BUTTON1_MASK, KeyEvent.CTRL_MASK);
        new Timeout("",50).sleep();
    }
    
    
    public boolean equals(Object el){
        if (el instanceof PatternOperator){
            return (this.pattern == ((PatternOperator)el).pattern);
        }else{
            return false;
        }
    }
    
    
    static {
        Timeouts.initDefault("PatternOperator.WaitPatternTime", WAIT_PATTERN_TIMEOUT);
    }
    
    
    
    
    public static class PatternComparator<C extends Pattern> implements Comparator<C>{
        public int compare(C o1, C o2){
            Point o1Loc = null;
            Point o2Loc = null;
            o1Loc = o1.createSelection().getBounds().getLocation();
            o2Loc = o2.createSelection().getBounds().getLocation();
            
            if (o1Loc.y>o2Loc.y){
                return 1;
            }else if (o1Loc.y == o2Loc.y){
                if(o1Loc.x > o2Loc.x){
                    return 1;
                } else{
                    return -1;
                }
            }else{
                return -1;
            }
        }
    }
    
    
    public static class PatternByTypeChooser implements IPatternChooser {
        
        private PatternTypes patternType = null;
        
        public PatternByTypeChooser(PatternTypes patternType){
            this.patternType = patternType;
        }
        
        
        public boolean checkPattern(Pattern pattern) {
            if (pattern==null){
                return false;
            }
            String inType=pattern.getOMReference().getClass().getName();
            
            JemmyProperties.getCurrentOutput().printLine("Found element of type "+inType+" while required "+patternType);
            
            if ( (patternType==null) || (!patternType.toString().equals(inType) && !(patternType == PatternTypes.ANY))){
                return false;
            }else{
                return true;
            }
        }
        
        public String getDescription( ) {
            return "Choose pattern by patterntype: " + patternType;
        }
    }
    
    public static class PatternByTypeAndNameChooser implements IPatternChooser {
        
        private PatternTypes patternType = null;
        private String patternName = null;
        private Operator.StringComparator comp = null;
        
        public PatternByTypeAndNameChooser(PatternTypes patternType, String patternName, Operator.StringComparator comparator){
            this.patternType = patternType;
            this.patternName = patternName;
            this.comp = comparator;
        }
        
        
        public boolean checkPattern(Pattern pattern) {
            if (pattern==null){
                return false;
            }
            String inType=pattern.getOMReference().getClass().getName();
            
            if ( (patternType==null) || (!patternType.toString().equals(inType) && !(patternType == PatternTypes.ANY))){
                return false;
            }else{
                //pattern type is ok
                //verifying name is ok:
                String name = pattern.getText();
                //JemmyProperties.getCurrentOutput().printLine("found element with name "+name +" required "+patternName);

                if (name==null || !comp.equals(name, patternName)){
                    return false;
                }else{
                    return true;
                }
            }
        }
        
        public String getDescription( ) {
            return "Choose pattern by patterntype and name: " + patternType;
        }
    }
    
    
    
    
    
    
    
}
