/*
 * The contents of this file are subject to the terms of the Common Development
 * and Distribution License (the License). You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the License at http://www.netbeans.org/cddl.html
 * or http://www.netbeans.org/cddl.txt.
 * 
 * When distributing Covered Code, include this CDDL Header Notice in each file
 * and include the License file at http://www.netbeans.org/cddl.txt.
 * If applicable, add the following below the CDDL Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * VisualElementOperator.java
 *
 */

package org.netbeans.orchlib.core.operators;

import java.awt.Component;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import org.netbeans.orchlib.exceptions.ElementNotFoundException;
import org.netbeans.jemmy.EventTool;
import org.netbeans.jemmy.JemmyProperties;
import org.netbeans.jemmy.Timeout;
import org.netbeans.jemmy.Timeouts;
import org.netbeans.jemmy.Waitable;
import org.netbeans.jemmy.Waiter;
import org.netbeans.jemmy.operators.JPopupMenuOperator;
import org.netbeans.jemmy.operators.Operator;
import org.netbeans.modules.bpel.design.geometry.FRectangle;
import org.netbeans.modules.bpel.design.model.connections.Connection;
import org.netbeans.modules.bpel.design.model.elements.VisualElement;
import org.netbeans.modules.bpel.design.model.patterns.CompositePattern;
import org.netbeans.modules.bpel.design.model.patterns.Pattern;

/**
 * This is the operator to work with visual elements available on the diagram
 * The operator allows to find elements on the diagram and also get some information about them
 *
 * @author Mikhail.Kondratyev@sun.com
 */
public class VisualElementOperator extends Operator implements IActionable{
    
    private PatternOperator patternOperator = null;
    private VisualElement visualElement = null;
    
    
    
    public final static long WAIT_ELEMENT_TIMEOUT = 60000;
    
    
    public VisualElementOperator(PatternOperator patternOper) throws ElementNotFoundException {
        this(patternOper, new ElementByTypeChooser(ElementTypes.ANY), 0);
    }
    
    
    public VisualElementOperator(PatternOperator patternOper, int index) throws ElementNotFoundException {
        this(patternOper, new ElementByTypeChooser(ElementTypes.ANY), index);
    }
    
    
    public VisualElementOperator(PatternOperator patternOper, ElementTypes type, int index) throws ElementNotFoundException {
        this(patternOper, new ElementByTypeChooser(type), index);
    }
    
    public VisualElementOperator(PatternOperator patternOper, ElementTypes type, String name, int index) throws ElementNotFoundException {
        this(patternOper, new ElementByTypeAndNameChooser(type, name, getDefaultStringComparator()), index);
    }
    
    public VisualElementOperator(PatternOperator patternOper, String name, int index) throws ElementNotFoundException {
        this(patternOper, new ElementByTypeAndNameChooser(ElementTypes.ANY, name, getDefaultStringComparator()), index);
    }
    
    
    public VisualElementOperator(PatternOperator patternOper, IElementChooser elementChooser, int index) throws ElementNotFoundException {
        this(patternOper, waitForElement(patternOper, elementChooser, index));
        
    }
    
    //TODO: it looks like this constructor is not needed and is the same as constructor below. 
    //needs to be removed later
    public VisualElementOperator(PatternOperator patternOper, VisualElement element) {        
        this.visualElement = element;
        DesignViewOperator view = patternOper.getDesignView();
        this.patternOperator = new PatternOperator(view, element.getPattern());
    }
    
    public VisualElementOperator(VisualElement element) {        
        this.visualElement = element;
        DesignViewOperator view = new DesignViewOperator(element.getPattern().getModel().getView());
        this.patternOperator = new PatternOperator(view, element.getPattern());
    }
    
    
    private static VisualElement findElement(PatternOperator patternOperator, IElementChooser elementChooser, int index) throws ElementNotFoundException{
        
        ArrayList<VisualElement> elementsFound = new ArrayList<VisualElement>();
        ArrayList<VisualElement> allElements = new ArrayList<VisualElement>();
        listChildElements(allElements, patternOperator.getPattern());
        
        for(int i=0;i<allElements.size();i++){
            if (elementChooser.checkElement(allElements.get(i))){
                elementsFound.add(allElements.get(i));
            }
        }
        
        //sorting found patterns
        ElementComparator<VisualElement> c = new ElementComparator<VisualElement>();
        VisualElement[] arr = (VisualElement[])elementsFound.toArray(new VisualElement[0]);
        Arrays.sort(arr,c);
        if (arr.length>index){
            return arr[index];
        }else{
            throw new ElementNotFoundException();
        }
    }
    
    static void listChildElements(ArrayList<VisualElement> elements, Pattern rootPattern){
        //adding elements of the root pattern
        Collection<VisualElement> nestedElements = rootPattern.getElements();
        for (Iterator<VisualElement> iter = nestedElements.iterator(); iter.hasNext();) {
            VisualElement nested = iter.next();
            //System.err.println("element type "+nested.getClass().getName());
            elements.add(nested);
        }
                
        //digging into the nested patterns
        if (rootPattern instanceof CompositePattern){            
            CompositePattern composite = (CompositePattern)rootPattern;
            Collection<Pattern> nestedPatterns = composite.getNestedPatterns();
            for (Iterator<Pattern> iter = nestedPatterns.iterator(); iter.hasNext();) {
                Pattern nested = iter.next();                
                listChildElements(elements, nested);
            }
        }
    }
    
    
    public static VisualElement waitForElement(final PatternOperator rootPattern, final IElementChooser elementChooser, final int index) {
        try{
            Waiter w = new Waiter(new Waitable() {
                public Object actionProduced(Object obj) {
                    try {
                        VisualElement go = findElement(rootPattern, elementChooser, index);
                        return go;
                    }catch(ElementNotFoundException e) {
                        return null;
                    }
                }
                public String getDescription() {
                    return("Wait for " + elementChooser.getDescription());
                }
            });
            Timeouts t = JemmyProperties.getCurrentTimeouts();
            t.setTimeout("Waiter.WaitingTime", t.getTimeout("VisualElementOperator.WaitElementTime"));
            return (VisualElement)w.waitAction(null);
            
        }catch(InterruptedException ie) {
            return null;
        }
    }
    
    
    
    /**
     * We must implement this method to extend Operator, but since
     * we are not working with Component this method always return null;
     * @return null
     */
    public Component getSource() {
        return null;
    }
    
    VisualElement getVisualElement(){
        return visualElement;
    }
    
    
   
    public DesignViewOperator getDesignView() {
        return patternOperator.getDesignView();
    }
    
    public PatternOperator getPattern(){
        return patternOperator;
    }    
    
    /**
     * returns pattern type
     */
    public String getTypeString() {
        return visualElement.getClass().getName();
    }
    
    
    /**
     *
     * @return bounding rect for element
     */
    public Rectangle getBoundingRect() {
        return getDesignView().convertRectToScreen((FRectangle) visualElement.getBounds().createRectangle());
    }
    
    
    public void clickActionPoint(int clickCount, int mouseBtn, int modifiers){
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, clickCount, mouseBtn, modifiers);
    }
    
    public void clickActionPoint(int clickCount){
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, clickCount, InputEvent.BUTTON1_MASK, 0);
    }
    
    
    public void clickForPopup() {
        makeVisible();
        getDesignView().clickMouse(getActionPoint().x,getActionPoint().y, 1, InputEvent.BUTTON3_MASK, 0);
    }
    
    public Point getActionPoint(){
        Rectangle elementRect = getBoundingRect();
        Point center = new Point((int)elementRect.getCenterX(),(int)elementRect.getCenterY());        
        return center;
    }
    
    public String getName(){
        return visualElement.getLabelText();
    }
    
    public void makeVisible(){
        Rectangle rect = null;
        //int inflate = 15;
        
        Rectangle boundingRect = getBoundingRect();
        //TODO: this comparison is kind of hack and should be work arounded if time permits
        if (boundingRect.getLocation().x<0 || boundingRect.getLocation().y<0){
            rect = new Rectangle(0,0, (int)boundingRect.getMaxX(), (int)boundingRect.getMaxY());
        }else{
            rect = boundingRect;
        }
        
        getDesignView().scrollToPoint(getActionPoint());
        getDesignView().scrollRectToVisible(rect);
        new EventTool().waitNoEvent(500);
    }
    
    
    
    public JPopupMenuOperator getPopup() {
        clickForPopup();
        return new JPopupMenuOperator();
    }
        
    
    public void select() {
        //TODO: wait selected should be added
        clickActionPoint(1, InputEvent.BUTTON1_MASK, 0);
        new Timeout("",50).sleep();
    }
    
    
    public boolean isSelected(){        
        return getDesignView().getDesignView().getSelectionModel().getSelectedPattern()==visualElement.getPattern();
    }
    
    
    public ArrayList<ConnectionOperator> getOutgoingConnections(){
        ArrayList<ConnectionOperator> conns = new ArrayList<ConnectionOperator>();
        List<Connection> list = visualElement.getOutcomingConnections();
        for(int i=0;i<list.size();i++){
            conns.add(new ConnectionOperator(list.get(i)));
        }
        return conns;
    }
    
    
    public void addToSelection() {
        new Timeout("",10).sleep();
        clickActionPoint(1, InputEvent.BUTTON1_MASK, KeyEvent.CTRL_MASK);
        new Timeout("",50).sleep();
    }
    
    
    public boolean equals(Object el){
        if (el instanceof VisualElementOperator){
            return (this.visualElement == ((VisualElementOperator)el).visualElement);
        }else{
            return false;
        }
    }
    
    
    static {
        Timeouts.initDefault("VisualElementOperator.WaitElementTime", WAIT_ELEMENT_TIMEOUT);
    }
    
    
    
    
    public static class ElementComparator<C extends VisualElement> implements Comparator<C>{
        public int compare(C o1, C o2){
            Point o1Loc = null;
            Point o2Loc = null;
            
            o1Loc = new Point((int)o1.getX(), (int)o1.getY());
            o2Loc = new Point((int)o2.getX(), (int)o2.getY());
                        
            if (o1Loc.y>o2Loc.y){
                return 1;
            }else if (o1Loc.y == o2Loc.y){
                if(o1Loc.x > o2Loc.y){
                    return 1;
                } else{
                    return -1;
                }
            }else{
                return -1;
            }
        }
    }
    
    
    public static class ElementByTypeChooser implements IElementChooser {
        
        private ElementTypes elementType = null;
        
        public ElementByTypeChooser(ElementTypes elementType){
            this.elementType = elementType;
        }        
        
        public boolean checkElement(VisualElement element) {
            if (element==null){
                return false;
            }
            String inType=element.getClass().getName();
            
            if ( (elementType==null) || (!elementType.toString().equals(inType) && !(elementType==ElementTypes.ANY))){
                return false;
            }else{
                return true;
            }
        }
        
        public String getDescription( ) {
            return "Choose element by visual element type: " + elementType;
        }
    }
    
    
    public static class ElementByTypeAndNameChooser implements IElementChooser {
        
        private ElementTypes elementType = null;
        private String name = null;
        private Operator.StringComparator comparator = null;
        
        public ElementByTypeAndNameChooser(ElementTypes elementType, String name, Operator.StringComparator comparator){
            this.elementType = elementType;
            this.name = name;
            this.comparator = comparator;
        }        
        
        public boolean checkElement(VisualElement element) {
            if (element==null){
                return false;
            }
            String inType=element.getClass().getName();
            
            if ( (elementType==null) || (!elementType.toString().equals(inType) && !(elementType==ElementTypes.ANY))){
                return false;
            }else{
                String elName = element.getLabelText();                
                if (elName==null || elName.equals("") || name==null || !comparator.equals(name, elName)){
                    return false;
                }else{
                    System.err.println("found"+elName+" type "+element.getClass().getName());
                    return true;
                }
            }
        }
        
        public String getDescription( ) {
            return "Choose element by visual element type and name: " + elementType+", "+name;
        }
    }

    protected PatternOperator getPatternOperator() {
        return patternOperator;
    }

    protected void setPatternOperator(PatternOperator patternOperator) {
        this.patternOperator = patternOperator;
    }
    
    
}
