/*
 * The contents of this file are subject to the terms of the Common Development
 * and Distribution License (the License). You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the License at http://www.netbeans.org/cddl.html
 * or http://www.netbeans.org/cddl.txt.
 * 
 * When distributing Covered Code, include this CDDL Header Notice in each file
 * and include the License file at http://www.netbeans.org/cddl.txt.
 * If applicable, add the following below the CDDL Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * Creator.java
 */

package org.netbeans.orchlib.ext.creator;

import java.io.File;
import java.util.ArrayList;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.netbeans.orchlib.core.operators.PatternOperator;
import org.netbeans.orchlib.core.operators.PatternTypes;
import org.netbeans.orchlib.exceptions.BadXMLSourceException;
import org.netbeans.orchlib.exceptions.ElementNotFoundException;
import org.netbeans.orchlib.ext.creator.actions.CreateAction;
import org.netbeans.orchlib.ext.creator.actions.CreateActionFactory;
import org.netbeans.orchlib.ext.creator.actions.CreatorUtil;
import org.netbeans.orchlib.ext.creator.actions.DummyNameAction;
import org.netbeans.orchlib.ext.creator.actions.LogAction;
import org.netbeans.orchlib.ext.creator.actions.LogActionFactory;
import org.netbeans.orchlib.ext.creator.actions.NameAction;
import org.netbeans.orchlib.ext.creator.actions.NameActionFactory;
import org.netbeans.orchlib.ext.creator.exceptions.CreationVerificationException;
import org.netbeans.orchlib.ext.creator.exceptions.NamingVerificationException;
import org.netbeans.orchlib.utils.TypeSpecificProperties;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * The Creator class is intended for building diagrams in designer and verifying the process
 * based on the BPEL documents supplied. The creator parses a BPEL document and creates it's
 * representation in designer. This allows to easily automatically test various combinations
 * of BPEL activities design and designing BPEL files of any complexity
 *
 * In addition to the main function - building diagram in designer, the Creator also may use logger 
 * to create test spec based on the performed actions
 *
 * @author Mikhail.Kondratyev@sun.com
 */
public class Creator {
    
    /** Creates a new instance of Creator */
    public Creator() {
    }
    
    
    /**
     * Runs the creator
     * @param filePath Path to the BPEL file to be parsed
     * @param processPattern Root BPEL pattern - we assume it should exist
     * @param logPath Path to write logs to
     * @throws java.lang.Exception 
     */
    public void run(String filePath, PatternOperator processPattern, String logPath) throws Exception{
        try{
            File logFile = new File(logPath);
            Logger.setOutFile(logFile);
            Logger log = Logger.instance();
            
            log.writeCase(filePath);
            
            DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            Document doc = builder.parse(new File(filePath));
            Element element = doc.getDocumentElement();
            
            ArrayList<Element> childElements = getChildElements(element);
            for(int i=0;i<childElements.size();i++){                
                buildPattern(childElements.get(i), processPattern);
            }
            
            log.writeCaseEnd();
        }catch(SAXException saxE){
            saxE.printStackTrace();
            throw new BadXMLSourceException();
        }
        
    }
    
    /**
     * Runs the creator
     * @param doc 
     * @param processPattern 
     * @throws orchlib.exceptions.ElementNotFoundException 
     * @throws orchlib.ext.creator.exceptions.CreationVerificationException 
     * @throws orchlib.ext.creator.exceptions.NamingVerificationException 
     */
    public void run(Document doc, PatternOperator processPattern) throws ElementNotFoundException, CreationVerificationException, NamingVerificationException{
        Element element = doc.getDocumentElement();
        
        ArrayList<Element> childElements = getChildElements(element);
        for(int i=0;i<childElements.size();i++){
            buildPattern(childElements.get(i), processPattern);
        }
    }
    
    /**
     * Creates a new activity
     * @param element Dom element that specifies what has to be created
     * @param parentPattern Parent pattern where new activity will be created
     * @throws orchlib.exceptions.ElementNotFoundException 
     * @throws orchlib.ext.creator.exceptions.CreationVerificationException 
     * @throws orchlib.ext.creator.exceptions.NamingVerificationException 
     */
    public void buildPattern(Element element, PatternOperator parentPattern) throws ElementNotFoundException, CreationVerificationException, NamingVerificationException{
        Logger log = Logger.instance();
        
        System.err.println("Element "+element.getTagName()+" in pattern "+parentPattern.getName());
        
        CreateActionFactory actionFactory = CreateActionFactory.getInstance();
        CreateAction createAction = actionFactory.createAction(parentPattern, element);
        NameAction nameAction = NameActionFactory.getInstance().createAction(parentPattern, element);
        LogAction logAction = LogActionFactory.getInstance().createAction(parentPattern, element);
        
        PatternTypes patternType = new CreatorUtil().getPatternTypeByTag(element.getTagName());
        if (createAction!=null){
            
            if (logAction!=null){
                logAction.performCreate(parentPattern, patternType);
            }else{
                String parentName = parentPattern.getName();
                if(parentName==null){
                    parentName = "newly created";
                }
                log.writeAction("Drag new "+TypeSpecificProperties.getCurrentPatternPaletteName(patternType)+" activity to the "+parentName+" activity from palette");
            }
            
            
            parentPattern = createAction.perform(parentPattern, patternType);
                        
            if (nameAction!=null){
                nameAction.perform(parentPattern, element.getAttribute("name"));
                if (nameAction instanceof DummyNameAction){
                    log.writeResult("New "+TypeSpecificProperties.getCurrentPatternPaletteName(patternType)+" pattern should be created at specified location.");
                }else{
                    log.writeAction("Set name "+element.getAttribute("name")+" for the new pattern.");
                    log.writeResult("New "+TypeSpecificProperties.getCurrentPatternPaletteName(patternType)+" pattern should be created at specified location. Name should be set");
                }
            }
        }else{
            System.err.println("Action was null for Element "+element.getTagName()+" in pattern "+parentPattern.getName());
        }
        
        ArrayList<Element> childElements = getChildElements(element);
        for(int i=0;i<childElements.size();i++){
            buildPattern(childElements.get(i), parentPattern);
        }
    }
    
    private ArrayList<Element> getChildElements(Element el){
        ArrayList<Element> elems = new ArrayList<Element>();
        NodeList list = el.getChildNodes();
        for(int i=0; i<list.getLength();i++){
            if (list.item(i) instanceof Element){
                Element child = (Element)list.item(i);
                elems.add(child);
            }
        }
        return elems;
    }
    
}
