/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.modules.povproject;

import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import org.netbeans.api.project.Project;
import org.netbeans.api.project.ProjectInformation;
import org.netbeans.modules.examples.api.povray.MainFileProvider;
import org.netbeans.modules.examples.api.povray.RendererService;
import org.netbeans.spi.project.ActionProvider;
import org.netbeans.spi.project.ProjectState;
import org.netbeans.spi.project.ui.LogicalViewProvider;
import org.openide.ErrorManager;
import org.openide.cookies.OpenCookie;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.util.Lookup;
import org.openide.util.RequestProcessor;
import org.openide.util.Utilities;
import org.openide.util.lookup.Lookups;
/**
 * Basic POV-Ray project support
 *
 * @author Tim Boudreau
 */
public final class PovrayProject implements Project {
    public static final String KEY_MAINFILE = "main.file";
    public static final String SCENES_DIR = "scenes"; //NOI18N
    public static final String IMAGES_DIR = "images"; //NOI18N

    private final FileObject projectDir;
    LogicalViewProvider logicalView = new PovrayLogicalView(this);
    private final ProjectState state;

    public PovrayProject(FileObject projectDir, ProjectState state) {
        this.projectDir = projectDir;
        this.state = state;
    }

    public FileObject getProjectDirectory() {
        return projectDir;
    }

    /**
     * Fetches the scenes folder, creating it if necessary.
     * Called from PovrayProject.saveProject().
     */
    FileObject getScenesFolder(boolean create) {
        FileObject result =
            projectDir.getFileObject(SCENES_DIR);

        if (result == null && create) {
            try {
                result = projectDir.createFolder (SCENES_DIR);
            } catch (IOException ioe) {
                ErrorManager.getDefault().notify(ioe);
            }
        }
        return result;
    }

    FileObject getImagesFolder(boolean create) {
        FileObject result =
            projectDir.getFileObject(IMAGES_DIR);
        if (result == null && create) {
            try {
                result = projectDir.createFolder (IMAGES_DIR);
            } catch (IOException ioe) {
                ErrorManager.getDefault().notify(ioe);
            }
        }
        return result;
    }

    private Lookup lkp;
    public Lookup getLookup() {
        if (lkp == null) {
            lkp = Lookups.fixed(new Object[] {
                this,  //project spec requires a project be in its own lookup
                state, //allow outside code to mark the project as needing saving
                new ActionProviderImpl(), //Provides standard actions like Build and Clean
                loadProperties(), //The project properties
                new Info(), //Project information implementation
                logicalView, //Logical view of project implementation
                new RendererServiceImpl(this), //Renderer Service Implementation
                new MainFileProviderImpl(this),
                new ViewServiceImpl (this),
            });
        }
        return lkp;
    }

    private Properties loadProperties() {
        FileObject fob = projectDir.getFileObject(PovProjectFactory.PROJECT_DIR +
            "/" + PovProjectFactory.PROJECT_PROPFILE);
        Properties properties = new NotifyProperties(state);
        if (fob != null) {
            try {
                properties.load(fob.getInputStream());
            } catch (Exception e) {
                ErrorManager.getDefault().notify(e);
            }
        }
        return properties;
    }

    private static class NotifyProperties extends Properties {
        private final ProjectState state;
        NotifyProperties (ProjectState state) {
            this.state = state;
        }

        public Object put(Object key, Object val) {
            Object result = super.put (key, val);
            if (((result == null) != (val == null)) || (result != null &&
                val != null && !val.equals(result))) {
                state.markModified();
            }
            return result;
        }
    }

    private final class ActionProviderImpl implements ActionProvider {
        public String[] getSupportedActions() {
            return new String[] { ActionProvider.COMMAND_BUILD,
                ActionProvider.COMMAND_CLEAN, ActionProvider.COMMAND_COMPILE_SINGLE };
        }

        public void invokeAction(String string, Lookup lookup) throws IllegalArgumentException {
            int idx = Arrays.asList (getSupportedActions()).indexOf (string);
            switch (idx) {
                case 0 : //build
                    final RendererService ren = (RendererService) getLookup().lookup(RendererService.class);
                    RequestProcessor.getDefault().post (new Runnable() {
                        public void run() {
                            FileObject image = ren.render();
                            if (image != null) {
                                DataObject dob;
                                try {
                                    dob = DataObject.find(image);
                                    OpenCookie open = (OpenCookie)
                                        dob.getNodeDelegate().getLookup().lookup(
                                        OpenCookie.class);
                                    if (open != null) {
                                        open.open();
                                    }
                                } catch (DataObjectNotFoundException ex) {
                                    ErrorManager.getDefault().notify(ex);
                                }
                            }
                        }
                    });
                    break;
                case 1 : //clean
                    FileObject fob = getImagesFolder(false);
                    if (fob != null) {
                        DataFolder fld = DataFolder.findFolder(fob);
                        for (Enumeration en=fld.children(); en.hasMoreElements();) {
                            DataObject ob = (DataObject) en.nextElement();
                            try {
                                ob.delete();
                            } catch (IOException ioe) {
                                ErrorManager.getDefault().notify(ioe);
                            }
                        }
                    }
                    break;
                case 2 : //compile-single
                    final DataObject ob = (DataObject) lookup.lookup (DataObject.class);
                    if (ob != null) {
                        final RendererService ren1 = (RendererService) getLookup().lookup(RendererService.class);
                        RequestProcessor.getDefault().post (new Runnable() {
                            public void run() {
                                if (ob.isValid()) { //Could theoretically change before we run
                                    ren1.render(ob.getPrimaryFile());
                                }
                            }
                        });
                    }
                    break;
                default :
                    throw new IllegalArgumentException (string);
            }
        }

        public boolean isActionEnabled(String string, Lookup lookup) throws IllegalArgumentException {
            int idx = Arrays.asList (getSupportedActions()).indexOf (string);
            boolean result;
            switch (idx) {
                case 0 :
                    result = true;
                    break;
                case 1 :
                    result = getImagesFolder(false) != null &&
                            getImagesFolder(false).getChildren().length > 0;
                    break;
                case 2 :
                    DataObject ob = (DataObject) lookup.lookup (DataObject.class);
                    if (ob != null) {
                        FileObject file = ob.getPrimaryFile();
                        result = "text/x-povray".equals(file.getMIMEType());
                    } else {
                        result = false;
                    }
                    break;
                default :
                    result = false;
            }
            return result;
        }
    }

    /** Implementation of project system's ProjectInformation class */
    private final class Info implements ProjectInformation {
        public Icon getIcon() {
            return new ImageIcon (Utilities.loadImage(
                "org/netbeans/examples/modules/povproject/resources/povicon.gif"));
        }

        public String getName() {
            return getProjectDirectory().getName();
        }

        public String getDisplayName() {
            return getName();
        }

        public void addPropertyChangeListener (PropertyChangeListener pcl) {
            //do nothing, won't change
        }

        public void removePropertyChangeListener (PropertyChangeListener pcl) {
            //do nothing, won't change
        }

        public Project getProject() {
            return PovrayProject.this;
        }
    }


}
