/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.iz;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.netbeans.iz.Query;
import org.netbeans.iz.QuerySet;
import org.netbeans.iz.XMLQueryHandler;

/**
 *
 * @author  jchalupa
 */
public class IZQuery {

    private static final String XML_FILE_HEADER =
        "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>";
    
    private static final String XML_ISSUES_URL = 
        "http://www.netbeans.org/issues/xml.cgi?id=";
    
    private static final String ISSUEZILLA_TAG_START =
        "<issuezilla version=\"2.11\" " +
        "urlbase=\"http://openide.netbeans.org/issues/\" " +
        "maintainer=\"owner@openide.netbeans.org\" " +
        "dtd_version=\"$Revision$\">";  

    private static final String DOCTYPE_START = "<!DOCTYPE issuezilla [";
    private static final String DOCTYPE_END = "]>";
    
    private static final String ISSUEZILLA_TAG_END = "</issuezilla>";
    
    private static final String PARAM_NO_ATTACHMENTS = "&show_attachments=false";
    
    private static final String DATA_FILE_PREFIX = "izq";
    private static final String MASTER_FILE_PREFIX = "izqm";
    private static final String OUTPUT_FILE_PREFIX = "izqo";
    private static final String XML_FILE_EXT = ".xml";

    // stylesheets
    private static final String PREPROCESS_XSL = "org/netbeans/iz/xslt/preprocess.xsl";
    private static final String NOOP_XSL = "org/netbeans/iz/xslt/noop.xsl";
    private static final String SIMPLE_TEXT_XSL = "org/netbeans/iz/xslt/simple-text.xsl";
    private static final String SIMPLE_HTML_XSL = "org/netbeans/iz/xslt/simple-html.xsl";

    private static final String NOOP_XSL_ID = "_none";
    private static final String SIMPLE_TEXT_XSL_ID = "_simple_text";
    private static final String SIMPLE_HTML_XSL_ID = "_simple_html";
    
    
    private static final int MAX_QUERY_SIZE = 50;

    private static final File tmpFolder = null;

    private boolean debug = false;
    private boolean runningInAnt = false;
    
    private String queryFile;
    private String transformFile;
    private String mergeTransformFile;
    private String outputFile;

    private String stoppersUmbrellaID;
    private String featureStoppersUmbrellaID;
    
    private QuerySet querySet;

    private List issueIdList = null;
    private List stoppersIdList = null;
    private List featureStoppersIdList = null;
    
    private List dataFiles = new ArrayList();
    private List stoppersDataFiles = new ArrayList();
    private List featureStoppersDataFiles = new ArrayList();


    private static Hashtable stylesheets;

    static {
        stylesheets = new Hashtable();
        
        stylesheets.put(NOOP_XSL_ID, NOOP_XSL);
        stylesheets.put(SIMPLE_TEXT_XSL_ID, SIMPLE_TEXT_XSL);
        stylesheets.put(SIMPLE_HTML_XSL_ID, SIMPLE_HTML_XSL);
    }
        
    
    /** Creates a new instance of IZQuery */
    public IZQuery() {
    }

    /** Sets the query file name */
    public void setQueryFile(String filename) {
        this.queryFile = filename;
    }

    /** Sets the tranform file name */
    public void setTransformFile(String filename) {
        this.transformFile = filename;
    }

    /** Sets the tranform file name */
    public void setMergeTransformFile(String filename) {
        this.mergeTransformFile = filename;
    }

    /** Sets the output file name */
    public void setOutputFile(String filename) {
        this.outputFile = filename;
    }

    /** Sets the stoppersUmbrellaID */
    public void setStoppersUmbrellaID(String id) {
    	this.stoppersUmbrellaID = id;
    }

    /** Sets the featureStoppersUmbrellaID */
    public void setFeatureStoppersUmbrellaID(String id) {
    	this.featureStoppersUmbrellaID = id;
    }
    
    /** Sets the debug flag */
    public void setDebug(boolean flag) {
        this.debug = flag;
    }
    
    
    /** 
     * Processes the XML query file. Returns a sorted list of IDs
     * of issues matching the criteria defined by the query set.
     */
    private void processQueryFile() throws Exception {

        // Create a JAXP SAXParserFactory and configure it
        SAXParserFactory spf = SAXParserFactory.newInstance();
        // spf.setValidating(validation);

        XMLReader xmlReader = null;

        // Create a JAXP SAXParser
        SAXParser saxParser = spf.newSAXParser();

        // Get the encapsulated SAX XMLReader
        xmlReader = saxParser.getXMLReader();

        XMLQueryHandler handler = new XMLQueryHandler();

        // Set the ContentHandler of the XMLReader
        xmlReader.setContentHandler(handler);

        // Tell the XMLReader to parse the XML document
        xmlReader.parse(new File(queryFile).toURL().toString());

        // process the xml files
        querySet = handler.getQuerySet();
        // execute the query set and return matching ids
        issueIdList = querySet.execute();
        // sort the list
        Collections.sort(issueIdList);
        if (debug) System.out.println("issueIdList="+issueIdList);
        
        if (stoppersUmbrellaID != null) {
        	stoppersIdList = getDependsOnIssues(stoppersUmbrellaID);
        	Collections.sort(stoppersIdList);
	        if (debug) System.out.println("stoppersIdList="+stoppersIdList);
        }
        
        if (featureStoppersUmbrellaID != null) {
        	featureStoppersIdList = getDependsOnIssues(featureStoppersUmbrellaID);
        	Collections.sort(featureStoppersIdList);
	        if (debug) System.out.println("featureStoppersIdList="+featureStoppersIdList);
        }
        issueIdList.removeAll(stoppersIdList);
        issueIdList.removeAll(featureStoppersIdList);
    }
    
    private ArrayList getDependsOnIssues(String id) throws Exception {
	URL url = new URL("http://www.netbeans.org/issues/xml.cgi?id=" + id);
        InputStream is = url.openStream();
        DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
	DocumentBuilder builder = builderFactory.newDocumentBuilder();
	Document document = builder.parse (is);
	
        ArrayList issues = new ArrayList();
        NodeList nodes = document.getElementsByTagName("issue");
        if (nodes == null || nodes.getLength() != 1) 
		return null;
	Element issue = (Element)(nodes.item(0));
	nodes = issue.getElementsByTagName("dependson");
	for (int i=0; i<nodes.getLength(); i++) {
		issues.add(Integer.valueOf(nodes.item(i).getChildNodes().item(0).getNodeValue()));
		
	}
        is.close();    	
        return issues;
    }
    
    private void collectData(List issueIdList, List dataFiles) throws Exception {

        // Instantiate a TransformerFactory.
        javax.xml.transform.TransformerFactory tFactory = 
                javax.xml.transform.TransformerFactory.newInstance();
        
        /* This getSystemResourceAsStream stopped working after switching to Ant 1.6.x.
         * Don't know why. getResource().openStream seeems to work fine. */

        // InputStream is = this.getClass().getClassLoader().getSystemResourceAsStream(PREPROCESS_XSL);
        InputStream is = this.getClass().getClassLoader().getResource(PREPROCESS_XSL).openStream();
        
        // Use the TransformerFactory to process the stylesheet Source and
        // generate a Transformer.
        javax.xml.transform.Transformer transformer = tFactory.newTransformer(
                new StreamSource(is));
       
        for (int i = 0; i < issueIdList.size() / MAX_QUERY_SIZE + 1; i++) {
            int from = i * MAX_QUERY_SIZE;
            int to = Math.min(from + MAX_QUERY_SIZE, issueIdList.size());
	    // if no ids left, leave the cycle immediately
	    if (from >= to)
		break;
            List sublst = issueIdList.subList(from, to);
            ListIterator it = sublst.listIterator();
            StringBuffer sb = new StringBuffer(XML_ISSUES_URL);
            while (it.hasNext()) {
                sb.append(it.next().toString());
                if (it.hasNext())
                    sb.append(",");
            }
            sb.append(PARAM_NO_ATTACHMENTS);

            if (debug)
                System.out.println("url = " + sb.toString());

            File tmpFile = File.createTempFile(DATA_FILE_PREFIX, XML_FILE_EXT, tmpFolder);

            dataFiles.add(tmpFile);

            URL url = new URL(sb.toString());

            HttpURLConnection con = null;
            BufferedOutputStream os = null;

            try {
                con = (HttpURLConnection)url.openConnection();
                os = new BufferedOutputStream(new FileOutputStream(tmpFile));

                // Use the Transformer to transform an XML Source and send the
                // output to a Result object.
                transformer.transform(
                    new javax.xml.transform.stream.StreamSource(con.getInputStream()), 
                    new javax.xml.transform.stream.StreamResult(os));
            }
            finally {
                if (os != null) os.close();
                if (con != null) con.disconnect();
            }
        }
        
    }

            
    private File createMasterFile(List dataFiles) throws Exception {

        PrintWriter pw = null;
        
        try {
            // consolidate responses into a single XML file
            File masterFile = File.createTempFile(MASTER_FILE_PREFIX, XML_FILE_EXT, tmpFolder);

            pw = new PrintWriter(new FileOutputStream(masterFile));
            pw.println(XML_FILE_HEADER);
            if (dataFiles.size() > 0) {
                pw.println(DOCTYPE_START);
                for (int i = 0; i < dataFiles.size(); i++) {
                    pw.println("  <!ENTITY file" + i + " SYSTEM \"" + ((File)dataFiles.get(i)).toURL().toString() + "\" >");
                }
                pw.println(DOCTYPE_END);
            }

            pw.println(ISSUEZILLA_TAG_START);

            for (int i = 0; i < dataFiles.size(); i++) {
                pw.println("  &file" + i + ";");
            }
            pw.println(ISSUEZILLA_TAG_END);
            return masterFile;

        } finally {
            if (pw != null) {
                pw.close();
            }
        }
    }
    
    
    private File transform(File masterFile) throws Exception {

        // Instantiate a TransformerFactory.
        javax.xml.transform.TransformerFactory tFactory = 
                javax.xml.transform.TransformerFactory.newInstance();

        // Use the TransformerFactory to process the stylesheet Source and
        // generate a Transformer.
        javax.xml.transform.Transformer transformer;

        // if transform file not specified, assume the _noop stylesheet
        if (transformFile == null || transformFile.equals(""))
            transformFile = NOOP_XSL_ID;
        
        String key;
        StreamSource source;
        // set the stylesheet...
        if ((key = (String)stylesheets.get(transformFile)) != null) {
            // ... either a default one

            // Workaround (see above)
            // source = new StreamSource(this.getClass().getClassLoader().getSystemResourceAsStream(key));
            source = new StreamSource(this.getClass().getClassLoader().getResource(key).openStream());
        }
        else {
            // ... or a one provided by the user
            source = new StreamSource(transformFile);
        }
        transformer = tFactory.newTransformer(source);
        
        // optional query-set attributes should be passed to the stylesheet as params
        Enumeration en = querySet.getParamNames();
        while (en.hasMoreElements()) {
            String name = (String)en.nextElement();
            transformer.setParameter(name, querySet.getParam(name));
        }
        
        OutputStream os = null;
        File outputFile = File.createTempFile(OUTPUT_FILE_PREFIX, XML_FILE_EXT, tmpFolder);

        try {
            os = new FileOutputStream(outputFile);

	    // Need to call StreamSource(systemID) here instead of StreamSource(InputStream)
	    // it seems that due to the Xerces-J bug #4026, Xerces cannot handle external
            // entities in a source which is passed as a stream. It seems to be fixed in
            // Xerces since version 1.4.4. Unfortunately, older versions of Xerces are
            // still being used often.
            transformer.transform(
                new javax.xml.transform.stream.StreamSource(masterFile.toURL().toString()), 
                new javax.xml.transform.stream.StreamResult(os));
        } finally {
            if (os != null) os.close();
        }
        return outputFile;
    }
    
    private void merge(File mainOutput, File stoppersOutput, File featureStoppersOutput) throws Exception {

        // Instantiate a TransformerFactory.
        javax.xml.transform.TransformerFactory tFactory = 
                javax.xml.transform.TransformerFactory.newInstance();

        // Use the TransformerFactory to process the stylesheet Source and
        // generate a Transformer.
        javax.xml.transform.Transformer transformer;

        // if transform file not specified, assume the _noop stylesheet
        if (mergeTransformFile == null || mergeTransformFile.equals(""))
            throw new IllegalArgumentException("Missing argument mergeTransformFile.");
        
        String key;
        StreamSource source;
        // set the stylesheet...
        if ((key = (String)stylesheets.get(mergeTransformFile)) != null) {
            // ... either a default one

            // Workaround (see above)
            // source = new StreamSource(this.getClass().getClassLoader().getSystemResourceAsStream(key));
            source = new StreamSource(this.getClass().getClassLoader().getResource(key).openStream());
        }
        else {
            // ... or a one provided by the user
            source = new StreamSource(mergeTransformFile);
        }
        transformer = tFactory.newTransformer(source);
        
        transformer.setParameter("file1", stoppersOutput);
        transformer.setParameter("file2", featureStoppersOutput);
        
        OutputStream os = null;

        try {
            if (outputFile != null) {
                os = new FileOutputStream(outputFile);
            }
            else {
                os = System.out;
            }

	    // Need to call StreamSource(systemID) here instead of StreamSource(InputStream)
	    // it seems that due to the Xerces-J bug #4026, Xerces cannot handle external
            // entities in a source which is passed as a stream. It seems to be fixed in
            // Xerces since version 1.4.4. Unfortunately, older versions of Xerces are
            // still being used often.
            transformer.transform(
                new javax.xml.transform.stream.StreamSource(mainOutput.toURL().toString()), 
                new javax.xml.transform.stream.StreamResult(os));
        } finally {
            if (os != null) os.close();
        }
    }
        
    
    public void execute() throws Exception {
    	
    	File masterFile = null;
        File stoppersMasterFile = null;
        File featureStoppersMasterFile = null;

        try {
            // process the query XML definition, execute the queries
            // and return the list of matching IDs
            processQueryFile();
        
            if (debug) {
                System.out.println("Found bugs: " + issueIdList.size());
            }

            // collect XML data
            collectData(issueIdList, dataFiles);
            if (stoppersIdList != null)
            	collectData(stoppersIdList, stoppersDataFiles);
            if (stoppersIdList != null)
            	collectData(featureStoppersIdList, featureStoppersDataFiles);
	    if (debug) {
	    	System.out.println("dataFiles="+dataFiles);
	    	System.out.println("stoppersDataFiles="+stoppersDataFiles);
	    	System.out.println("featureStoppersDataFiles="+featureStoppersDataFiles);
	    }

            // create the master XML file...
            masterFile = createMasterFile(dataFiles);
            if (stoppersIdList != null)
            	stoppersMasterFile = createMasterFile(stoppersDataFiles);
            if (stoppersIdList != null)
            	featureStoppersMasterFile = createMasterFile(featureStoppersDataFiles);
            
            // ... and transform it using the provided XSL stylesheet
            File mainOutput = transform(masterFile);
            File stoppersOutput = null;
            if (stoppersMasterFile != null)
            	stoppersOutput = transform(stoppersMasterFile);
            File featureStoppersOutput = null;
            if (featureStoppersMasterFile != null)
            	featureStoppersOutput = transform(featureStoppersMasterFile);
            	
            merge(mainOutput, stoppersOutput, featureStoppersOutput);
            	
        }
        finally {
            // clean the temporary files (unless in debug mode)
            if (!debug) {
                for (int i = 0; i < dataFiles.size(); i++) {
                    ((File)dataFiles.get(i)).delete();
                }
                for (int i = 0; i < stoppersDataFiles.size(); i++) {
                    ((File)stoppersDataFiles.get(i)).delete();
                }
                for (int i = 0; i < featureStoppersDataFiles.size(); i++) {
                    ((File)featureStoppersDataFiles.get(i)).delete();
                }
                if (masterFile != null) {
                    masterFile.delete();
                }
                if (stoppersMasterFile != null) {
                    stoppersMasterFile.delete();
                }
                if (featureStoppersMasterFile != null) {
                    featureStoppersMasterFile.delete();
                }
            }
        }
        
                
    }

    
    private boolean processOptions(String[] args) {
        
        for (int i = 0; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-q") || args[i].equalsIgnoreCase("-query")) {
                if (i + 1 < args.length) {
                    setQueryFile(args[++i]);
                }
                else {
                    System.out.println("Query file name missing.");
                    return false;
                }
            }
            else if (args[i].equalsIgnoreCase("-t") || args[i].equalsIgnoreCase("-transform")) {
                if (i + 1 < args.length) {
                    setTransformFile(args[++i]);
                }
                else {
                    System.out.println("Transform file name missing.");
                    return false;
                }
            }
            else if (args[i].equalsIgnoreCase("-o") || args[i].equalsIgnoreCase("-out") || args[i].equalsIgnoreCase("-output")) {
                if (i + 1 < args.length) {
                    setOutputFile(args[++i]);
                }
                else {
                    System.out.println("Output file name missing.");
                    return false;
                }
            }
            else if (args[i].equalsIgnoreCase("-d") || args[i].equalsIgnoreCase("-debug")) {
                setDebug(true);
            }
            else {
                System.out.println("Invalid argument: " + args[i]);
                return false;
            }
        }
        if (queryFile == null) {
            System.out.println("Query file name argument missing.");
            return false;
        }
        
        // everything seems to be ok
        return true;        
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {

        IZQuery self = new IZQuery();
        
        if (!self.processOptions(args)) {
            System.exit(-1);
        }
        
        try {
            self.execute();
        }
        catch (Exception ex) {
            System.err.println(ex.getMessage());
            System.exit(-1);
        }
    }

    
    
}
