/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.testspeceditor.undo;

import java.util.LinkedList;
import org.netbeans.testspeceditor.events.Bean;

/**
 *
 * @author ehucka
 */
public class UndoManager {
    
    private static UndoManager instance = null;
    
    LinkedList<UndoAction> undoStack;
    LinkedList<UndoAction> redoStack;
    int maxUndoDeep = 50;
    boolean undoing=false;
    AtomicAction atomic=null;
    boolean logproperties=true;
    
    Bean oldSource = null;
    String oldPropertyName = null;
    
    Object mutex = new Object();
    
     
    public UndoManager() {
        undoStack=new LinkedList();
        redoStack=new LinkedList();
    }
    
    public static UndoManager getInstance() {
        if (instance == null) {
            instance = new UndoManager();
        }
        return instance;
    }
    
    public void startAtomic() {
        if (atomic == null) {
            atomic=new AtomicAction();
        } else {
            throw new IllegalStateException("Cannot start atomic when it is already started.");
        }
    }
    
    public void endAtomic() {
        if (atomic != null) {
            if (!atomic.isEmpty())
                undoStack.addFirst(atomic);
            atomic=null;
        } else {
            throw new IllegalStateException("Cannot stop atomic when it is not started yet.");
        }
    }
    
    public void setLogProperties(boolean value) {
        logproperties=value;
    }
    
    public void childrenChanged(Bean source, Bean child, int oldIndex, int newIndex) {
        if (!undoing) {
            UndoAction act=new BeanStructureUndoAction(source, child, oldIndex, newIndex);
            packPropertiesEvents();
            addNewAction(act);
        }
    }
    
    public void propertyChanged(Bean source, String propertyName, Object oldValue, Object newValue) {
        if (!undoing && logproperties) {
            if (oldValue == newValue || oldValue != null && oldValue.equals(newValue)) {
                return;
            }
            UndoAction act=new PropertyChangedUndoAction(source, propertyName, oldValue, newValue);
            //pack after new property of new source is added
            if (oldSource != null && (oldSource != source || !propertyName.equals(oldPropertyName))) {
                packPropertiesEvents();
            }
            if (oldSource == null) {
                oldSource=source;
                oldPropertyName=propertyName;
            }
            addNewAction(act);
        }
    }
    
    private void addNewAction(final UndoAction act) {
        if (atomic == null) {
            undoStack.addFirst(act);
            while (undoStack.size() > maxUndoDeep) {
                undoStack.removeLast();
            }
        } else {
            atomic.addAction(act);
        }
        redoStack.clear();
    }
    
    protected void packPropertiesEvents() {
        if (!undoStack.isEmpty() && undoStack.getFirst() instanceof PropertyChangedUndoAction &&
                atomic == null && oldSource != null) {
            String oldValue, newValue;
            PropertyChangedUndoAction act=(PropertyChangedUndoAction)undoStack.getFirst();
            PropertyChangedUndoAction old=act;
            newValue=(String)act.newValue;
            oldValue=(String)act.oldValue;
            
            while (act != null && act.source == oldSource && act.propertyName.equals(oldPropertyName)) {
                act=(PropertyChangedUndoAction)undoStack.removeFirst();
                oldValue=(String)act.oldValue;
                if (!undoStack.isEmpty() && undoStack.getFirst() instanceof PropertyChangedUndoAction)
                    act=(PropertyChangedUndoAction)undoStack.getFirst();
                else
                    act=null;
            }
            if (old != act) {
                act=new PropertyChangedUndoAction(oldSource, oldPropertyName, oldValue, newValue);
                undoStack.addFirst(act);
            }
            oldSource=null;
            oldPropertyName=null;
        }
    }
    
    public boolean canUndo() {
        return (!undoStack.isEmpty());
    }
    
    public boolean canRedo() {
        return (!redoStack.isEmpty());
    }
    
    public void flush() {
        undoStack.clear();
        redoStack.clear();
    }
    
    public void undo() {
        if (!undoStack.isEmpty()) {
            try {
                UndoAction ua=undoStack.removeFirst();
                undoing=true;
                ua.undo();
                redoStack.addFirst(ua);
                undoing=false;
            } catch (UndoException ex) {
                ex.printStackTrace();
            }
        }
    }
    
    public void redo() {
        if (!redoStack.isEmpty()) {
            try {
                UndoAction ua=redoStack.removeFirst();
                undoing=true;
                ua.redo();
                undoStack.addFirst(ua);
                undoing=false;
            } catch (UndoException ex) {
                ex.printStackTrace();
            }
        }
    }
    
    protected void writeUndoStack() {
        System.out.println("Stack:");
        for (UndoAction act:undoStack) {
            System.out.println(act);
        }
    }
}
