/*
                    Sun Public License Notice
 
The contents of this file are subject to the Sun Public License Version
1.0 (the "License"). You may not use this file except in compliance with
the License. A copy of the License is available at http://www.sun.com/
 
The Original Code is the Regsup module.
The Initial Developer of the Original Code is Petr Slegr.
Portions created by Petr Slegr are Copyright (C) 2002.
All Rights Reserved.
 
Contributor(s): Petr Slegr.
 */

package org.netbeans.modules.regsup.logic;

import java.util.*;
import java.beans.*;
import org.netbeans.modules.regsup.logic.aut2expr.*;
import org.netbeans.modules.regsup.util.Util;

/**
 *
 * @author  Petr Slegr
 * @version 24.6.01/2
 */
public class LogicAutomaton /* implements PropertyChangeListener */ {
    
    public static final String PROP_NODE_COUNT = "nodeCount";
    
    public static final String PROP_EDGE_COUNT = "edgeCount";
    
    public static final String PROP_DETERMIN = "determin";
    
    /** Property whether this automaton should be DFA or NFA.
     */
    public static final String PROP_SHOULD_DETERMIN = "shouldDetermin";
    
    /** Property whether this automaton should have only one starting node.
     * Only NFAs can have more than one.
     */
    public static final String PROP_MORE_STARTING = "shouldOneStart";
    
    /** Property whether this automaton can contain edge with empty string
     * It has sense only for NFA.
     */
    public static final String PROP_EMPTY_ALLOWED = "emptyAllowed";
    
    private ArrayList nodes = new ArrayList ();
    
    private ArrayList finalNodesCache = new ArrayList ();
    
    private ArrayList startingNodesCache = new ArrayList ();
    
    /** null means that cache is invalid and automaton has to be checked again.
     * Zero size means that automaton is deterministic.
     */
    private ArrayList nonDeterminNodesCache = new ArrayList ();
    
    // !!!! rename
    private LogicAutomatonNode.NodeID idManager;
    
    /** Is this automaton currently deterministic
     */
    private boolean determin;
    
    /** Should be this automaton deterministic?
     */
    private boolean shouldDetermin = true;
    
    /** Should have this automaton only one or more starting nodes?
     */
    // !!!! pokud bude false a shouldDetermin == true, tak hazet vyjimku
    private boolean shouldOneStart = true;
    
    // !!!!!
    private RexElement rexElement;
    
    private PropertyChangeSupport propChangeSup = new PropertyChangeSupport (this);
    
    /** Singleton for listening on nodes
     */
    private NodeListenPoint nodeListener = new NodeListenPoint ();
    
    private static final int EDGE_COUNT_INVALID = -1;
    
    private int edgeCountCache = EDGE_COUNT_INVALID;
    
    /** Creates new LogicAutomaton */
    public LogicAutomaton () {
        idManager = new LogicAutomatonNode.NodeID ();
    }
    
    public void addPropertyChangeListener (PropertyChangeListener pcl) {
        propChangeSup.addPropertyChangeListener (pcl);
    }
    
    public void addPropertyChangeListener (String propName, PropertyChangeListener pcl) {
        propChangeSup.addPropertyChangeListener (propName, pcl);
    }
    
    public void removePropertyChangeListener (PropertyChangeListener pcl) {
        propChangeSup.removePropertyChangeListener (pcl);
    }
    
    public void removePropertyChangeListener (String propName, PropertyChangeListener pcl) {
        propChangeSup.removePropertyChangeListener (propName, pcl);
    }
    
    /**
     * @return true if this LogicAutomaton is ok
     */
    public boolean isResolved () {
        return rexElement != null ? true : false;
    }
    
    /** Returns regular expression that is represented by this automaton.
     * @return resolved expression if this LogicAutomaton is ok
     * @return null otherwise
     */
    public RexElement getExpression () {
        return rexElement;
    }
    
    /**
     * @return Set of all nodes in this automaton.
     */
    public Set getNodes () {
        return new HashSet (nodes);
    }
    
    /**
     */
    public void setShouldDetermin (boolean dfa) {
        boolean old = shouldDetermin;
        shouldDetermin = dfa;
        if (old != shouldDetermin) {
            // !!!! firovat eventu
        }
    }
    
    // !!!! prejmenovat na is
    public boolean getShouldDetermin () {
        return shouldDetermin;
    }
    
    /**
     * Creates new node and adds it into this automaton
     * @sidefect
     */
    public LogicAutomatonNode createNode () {
        LogicAutomatonNode node = new LogicAutomatonNode ();
        // !!!! zrusit, nepouziva se
        node.included = false;
        addNode (node);
        return node;
    }
    
    /**
     */
    public CompoundLogicAutomatonNode createCompoundNode (RexElement rex) {
        CompoundLogicAutomatonNode node = new CompoundLogicAutomatonNode (rex);
        node.included = false;
        addNode (node);
        return node;
    }
    
    /**
     */
    public CompoundLogicAutomatonNode createCompoundNode (LogicAutomaton automaton) {
        CompoundLogicAutomatonNode node = new CompoundLogicAutomatonNode (automaton);
        addNode (node);
        node.included = false;
        return node;
    }
    
    /** NOTE: NOT for API usage!
     */
    public void addNode (LogicAutomatonNode node) {
        boolean res = nodes.add (node);
        if (!res)
            throw new IllegalStateException ("Node already present: " + node);
        node.addPropertyChangeListener (nodeListener);
        node.setID (idManager.next ());
        //
        // was bug - missing
        propChangeSup.firePropertyChange (PROP_NODE_COUNT, null, null);
        if (! node.isIsolated ()) {
            propChangeSup.firePropertyChange (PROP_EDGE_COUNT, null, null);
            // ! possible to store into cache like bellow
        }
        if (node.isFinalNode ()) {
            if (finalNodesCache != null) {
                finalNodesCache.add (node);
            }
        }
        if (node.isStarting ()) {
            if (startingNodesCache != null) {
                startingNodesCache.add (node);
            }
        }
        if (! node.isDetermin ()) {
            if (nonDeterminNodesCache != null) {
                nonDeterminNodesCache.add (node);
            }
        }
    }
    
    /**
     */
    public void removeNode (LogicAutomatonNode node) {
        boolean res = nodes.remove (node);
        if (!res)
            throw new IllegalStateException ("Node not present: " + node);
        //
        if (node.isFinalNode ()) {
            if (finalNodesCache != null) {
                finalNodesCache.remove (node);
            }
        }
        if (node.isStarting ()) {
            if (startingNodesCache != null) {
                startingNodesCache.remove (node);
            }
        }
        if (! node.isDetermin ()) {
            if (nonDeterminNodesCache != null) {
                nonDeterminNodesCache.remove (node);
            }
        }
        node.removePropertyChangeListener (nodeListener);
        propChangeSup.firePropertyChange (PROP_NODE_COUNT, null, null);
        // was bug - missing
        if (! node.isIsolated ()) {
            propChangeSup.firePropertyChange (PROP_EDGE_COUNT, null, null);
            // ! cache
        }
    }
    
    /** Returns starting nodes.
     * @return Set of such nodes
     */
    public Set getStartingNodes () {
        if (startingNodesCache == null) {
            startingNodesCache = new ArrayList ();
            Iterator it = nodes.iterator ();
            while (it.hasNext ()) {
                LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                if (node.isStarting ()) {
                    startingNodesCache.add (node);
                }
            }
        }
        return new HashSet (startingNodesCache);
    }
    
    /** Looks for final nodes.
     * @return Set of such nodes
     */
    public Set getFinalNodes () {
        if (finalNodesCache == null) {
            finalNodesCache = new ArrayList ();
            Iterator it = nodes.iterator ();
            while (it.hasNext ()) {
                LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                if (node.isFinalNode ()) {
                    finalNodesCache.add (node);
                }
            }
        }
        return new HashSet (finalNodesCache);
    }
    
    
    /** Looks for isolated nodes
     * @return Set of such nodes
     */
    // !!! use cache
    public Set getIsolatedNodes () {
        Set res = new HashSet ();
        Iterator it = nodes.iterator ();
        while (it.hasNext ()) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            if (node.isIsolated ()) {
                res.add (node);
            }
        }
        return res;
    }
    
    /** Looks for nodes which are not final and have not input edge or
     * are not starting and have not output edge.
     * @return Set of such nodes
     */
    private Set findBadNodes () {
        Set res = new HashSet ();
        Iterator it = nodes.iterator ();
        while (it.hasNext ()) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            if (node.nextCount () == 0 && ! node.isFinalNode ()) {
                //System.out.println ("not final");
                res.add (node);
            }
            if (node.prevCount () == 0 && ! node.isStarting ()) {
                //System.out.println ("not starting");
                res.add (node);
            }
        }
        return res;
    }
    
    /** Looks for non-deterministic nodes.
     * @return Set of such nodes
     */
    public Set getNonDeterminNodes () {
        if (nonDeterminNodesCache == null) {
            nonDeterminNodesCache = new ArrayList ();
            Iterator it = nodes.iterator ();
            while (it.hasNext ()) {
                LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                if (! node.isDetermin ()) {
                    System.out.println("nondet node: " + node);
                    nonDeterminNodesCache.add (node);
                }
            }
        }
        return new HashSet (nonDeterminNodesCache);
    }
    
    /**
     */
    public int getNodeCount () {
        return nodes.size ();
    }
    
    /**
     */
    public int getStartingNodeCount () {
        return getStartingNodes ().size ();
    }
    
    /**
     */
    public int getFinalNodeCount () {
        return getFinalNodes ().size ();
    }
    
    /**
     */
    public int getIsolatedNodeCount () {
        return getIsolatedNodes ().size ();
    }
    
    /** Returns number of edges in automaton (does NOT include subautomatons).
     * Does not include self-edges
     */
    public int getEdgeCount () {
        if (edgeCountCache == EDGE_COUNT_INVALID) {
            edgeCountCache = 0;
            for (Iterator it = nodes.iterator (); it.hasNext (); ) {
                LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                edgeCountCache += node.getNextNodes ().size ();
            }
        }
        return edgeCountCache;
    }
    
    /**
     * Not implemented yet.
     */
    public Set getNodesWithSelfEdge () {
        throw new UnsupportedOperationException ();
    }
    
    /** If this automaton has less or more than one starting node throws IllegalStateException
     * @return the one starting node
     */
    // !!!! radsi vracet Set a user at testuje jeji velikost
    /*
    public LogicAutomatonNode getStartingNode () {
        Set starts = getStartingNodes ();
        if (starts.size () != 1) {
            throw new IllegalStateException ("Automaton should have only 1 starting node but has: "
            + starts.size () + " ones");
        }
        return (LogicAutomatonNode) starts.toArray ()[0];
    }
     */
    
    public String toDebugString () {
        StringBuffer buf = new StringBuffer ();
        buf.append ("LogicAutomaton: ");
        buf.append (hashCode ());
        buf.append (" nodeCount: ");
        buf.append (getNodeCount ());
        /*
        System.out.print ("LogicAutomaton: " + + nodeCount: " + );
        System.out.print (" isolatedNodeCount: " + isolatedNodeCount);
        System.out.print (" finalNodeCount: " + finalNodeCount);
        System.out.println (" starting#: " + getStartingNodeCount ());
        System.out.println (" CORRECT: " + isCorrect ());
        System.out.println ();
         */
        //System.out.println ("EXPR: " + getExpression ());
        return buf.toString ();
    }
    
    public RegEquationSet toEquationSet () {
        RegEquationSet res = new RegEquationSet ();
        Iterator it = nodes.iterator ();
        while (it.hasNext ()) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            //System.out.println ("processed node: " + node);
            RegEquation equ = new RegEquation (node.id, node.getInnerElement (),
            node.isStarting (), node.isFinalNode ());
            res.addEquation (equ);
            Iterator edg = node.getInputEdges ().iterator ();
            while (edg.hasNext ()) {
                LogicAutomatonEdge edge = (LogicAutomatonEdge) edg.next ();
                //System.out.println ("processed edge: " + edge);
                LogicAutomatonNode from = edge.getSource ();
                equ.addMember (from.id, edge.getEdgeValue ());
            }
            if (node.hasSelfEdge ()) {
                equ.addMember (node.id, node.getSelfEdge ().getEdgeValue ());
            }
        }
        return res;
    }
    
    /** Utility method
     */
    private static Set getFinalNodes (Set set) {
        Set res = new HashSet ();
        Iterator it = set.iterator ();
        while (it.hasNext ()) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            if (node.isFinalNode ()) {
                res.add (node);
            }
        }
        return res;
    }
    
    /** BFS
     */
    private static Set findFlowFrom (Set set) {
        ArrayList closed = new ArrayList ();
        ArrayList open = new ArrayList (set);
        while (open.size () != 0) {
            LogicAutomatonNode node = (LogicAutomatonNode) open.get (0);
            // ! not necessary to pay attention to self-edges
            Set nextNodes = node.getNextNodes ();
            for (Iterator it = nextNodes.iterator(); it.hasNext(); ) {
                Object o = it.next ();
                if (! closed.contains (o)) {
                    open.add (o);
                }
            }
            open.remove (node);
            closed.add (node);
        }
        return new HashSet (closed);
    }
    
    /** Finds all nodes available from starting nodes and returns them (including starting)
     */
    private Set findFlow () {
        return findFlowFrom (getStartingNodes ());
    }
    
    /**
     */
    public ResolveResult resolve () {
        rexElement = null; // cleares cache
        ResolveResult res = new ResolveResult ();
        //
        if (getStartingNodeCount () < 1) {
            res.addProblem (new DirectResolveError.MissingStartingNode ());
        }
        else if (shouldOneStart || shouldDetermin) {
            if (getStartingNodeCount () > 1 && shouldOneStart) {
                res.addProblem (new DirectResolveError.TooManyStartingNodes ());
                // !!!??? co vyrobit lokalni errory pro vsechny startovni nody?
                // ... vede na stromovou strukturu...jeden error ma vice pricin
                // spis jednu entry ale multiselection v editoru
            }
        }
        //
        /*
        if (getFinalNodeCount () < 1) {
            res.addProblem (new DirectResolveError.MissingFinalNode ());
        }
         */
        //
        Set reachableNodes = findFlow ();
        Set finalFlowReachableNodes = getFinalNodes (reachableNodes);
        System.out.println ("finalFlowReachableNodes: " + finalFlowReachableNodes);
        if (finalFlowReachableNodes.size () < 1) {
            res.addProblem (new DirectResolveError.MissingFinalNode ());
        }
        //
        Set unReachableNodes = new HashSet (nodes);
        unReachableNodes.removeAll (reachableNodes);
        // ! fatal-test must be performed always
        // !!!! fatal-test delat samostatne a vzdy
        for (Iterator it = unReachableNodes.iterator (); it.hasNext (); ) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            // fatal-test
            if (! node.isStarting () && node.isFinalNode ()) {
                DirectResolveProblem problem = new DirectResolveProblem.UnreachableNode (node, true);
                problem.setName ("Unreachable final node");
                res.addProblem (problem);
                // was bug: isolated node that is not starting and is final causes exception
                //(no solution) -> fatal
            }
            else if (! node.isFinalNode ()) {
                // non-fatal
                res.addProblem (new DirectResolveProblem.UnreachableNode (node, false));
            }
        }
        // ! not necessary test
        for (Iterator it = reachableNodes.iterator (); it.hasNext (); ) {
            LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
            if (! node.isFinalNode () && node.nextCount () == 0) {
                res.addProblem (new DirectResolveProblem.DullNode (node, false));
            }
        }
        /*
        Set isolatedNodes = getIsolatedNodes ();
        System.out.println ("isolatedNodes: " + isolatedNodes + " within: " + this);
        Iterator iter = isolatedNodes.iterator ();
        while (iter.hasNext ()) {
            LogicAutomatonNode node = (LogicAutomatonNode) iter.next ();
            if (! node.isStarting () && node.isFinalNode ()) {
                res.addProblem (new DirectResolveProblem.IsolatedNode (node, true));
                // was bug: isolated node that is not starting and is final causes exception
                //(no solution) -> fatal
            }
            else if (! node.isFinalNode ()) {
                // non-fatal
                res.addProblem (new DirectResolveProblem.IsolatedNode (node, false));
            }
        }
        //
        // !!!! rozlisit blank a unreachable
        Set badNodes = findBadNodes ();
        badNodes.removeAll (isolatedNodes); // isolated node should not be two times errorneous
        if (badNodes.size () != 0) {
            Iterator it = badNodes.iterator ();
            while (it.hasNext ()) {
                LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                res.addProblem (new DirectResolveProblem.BlankNode (node, false));
            }
        }
         */
        //
        if (shouldDetermin) {
            Set nonDetNodes = getNonDeterminNodes ();
            System.out.println ("nonDetNodes: " + nonDetNodes);
            if (nonDetNodes.size () != 0) {
                Iterator it = nonDetNodes.iterator ();
                while (it.hasNext ()) {
                    LogicAutomatonNode node = (LogicAutomatonNode) it.next ();
                    // non-deterministic node is error if determinism is desired
                    res.addProblem (new DirectResolveProblem.NonDeterminNode (node, true));
                }
                determin = false;
            } else {
                determin = true;
            }
        }
        // recursively down
        for (int i = 0; i < nodes.size (); i++) {
            LogicAutomatonNode node = (LogicAutomatonNode) nodes.get (i);
            if (node instanceof CompoundLogicAutomatonNode) {
                CompoundLogicAutomatonNode compound = (CompoundLogicAutomatonNode) node;
                ResolveResult innerRes = compound.resolve ();
                if (! innerRes.isOk ()) {
                    Set innProblems = innerRes.getProblems ();
                    Iterator it = innProblems.iterator ();
                    while (it.hasNext ()) {
                        ResolveProblem problem = (ResolveProblem) it.next ();
                        res.addProblem (new ProxyResolveProblem (compound, problem));
                    }
                }
            }
        }
        //
        if (res.errorsCount () > 0) { // was BUG - don't call isOk() here
            return res;
        }
        //
        RegEquationSet eqSet = toEquationSet ();
        RexElement rexRes = eqSet.solve ();
        System.out.println ("solution: " + rexRes);
        res.setResult (rexRes);
        //
        rexElement = rexRes; // result chache
        return res;
    }
    
    /**
     */
    public boolean isDetermin () {
        System.out.println("wwwwwwwwwwwwwwwwwwwww");
        return getNonDeterminNodes ().size () == 0;
    }

    /**
     */
    public String toString () {
        StringBuffer buf = new StringBuffer ();
        buf.append ("nodes: " + nodes.toString ());
        buf.append ("\n");
        //buf.append ("\tstarting nodes: " + startNodes.toString ());
        //buf.append ("\tfinal nodes: " + toString ());
        return buf.toString ();
    }
    
    public static void main (String[] args) {
    }
    
    ////////////////////////////////////////////////////////////////////////////
    private class NodeListenPoint implements PropertyChangeListener {
        
        /**
         * This method gets called when a bound property is changed.
         * @param evt A PropertyChangeEvent object describing the event source
         *  	and the property that has changed.
         */
        public void propertyChange (PropertyChangeEvent evt) {
            System.out.println("propName: " + evt.getPropertyName ());
            System.out.println("propSource: " + evt.getSource ());
            if (evt.getPropertyName ().equals (PROP_DETERMIN)) {
                // ! was bug: cache was not cleared
                nonDeterminNodesCache = null; // chache invalidation
                System.out.println("vvvvvvvvvvvvvv");
                propChangeSup.firePropertyChange (PROP_DETERMIN, null, null);
            }
            else if (evt.getPropertyName ().equals (PROP_EDGE_COUNT)) {
                // cache reset
                edgeCountCache = EDGE_COUNT_INVALID;
                propChangeSup.firePropertyChange (PROP_EDGE_COUNT, null, null);
            }
            else if (evt.getPropertyName ().equals (LogicAutomatonNode.PROP_STARTING)) {
                startingNodesCache = null;
                // it is also possible to fire event about this
            }
            else if (evt.getPropertyName ().equals (LogicAutomatonNode.PROP_FINAL_NODE)) {
                finalNodesCache = null;
                // it is also possible to fire event about this
            }
            
        }
        
    }
    
    
}
